/**
 * fs/directory.h
 * (c) 2006-2008 Murat Deligonul
 */

#ifndef __FS_DIRECTORY_H
#define __FS_DIRECTORY_H

#include <vector>
#include <deque>
#include <algorithm>
#include <ostream>
#include <ctime>
#include <sys/types.h>
#include "util/hash.h"
#include "util/tracked_object.h"
#include "fs/entry.h"

namespace fs {

/**
 * Represents a directory in the on-disk file library.
 */
class directory : 
#ifdef __DEBUG__
	public util::tracked_object<directory> {
#else
	public util::counted_object<directory> {
#endif
public:
	/**
	  * Normal C-string hasher with case-sensitive key comparison.
	  */
	typedef util::hash_table<const char *, file_entry *, 
					util::default_hasher<const char *>,
					util::fixed_cstring_predicate<strcmp> >	hash_table_t;

private:
	directory * _parent;		// parent directory object, if there is one
	char * dbfile;			// cached, full path to database file
	char * a_path;			// absolute path to this directory (i.e. on disk)
	const char * r_path;		// relative path to this directory (i.e. starting from file library root)
	const char * _name;		// name of this directory
	file_entry * _entry;		// file entry

	int fd;				// file descriptor of open database file
	bool batch_mode;		// whether we are executing multiple operations in one access.

	time_t last_mod;		// when database file was last modified, according to our records
	off_t  last_size;		// last known size of the database file	

	hash_table_t			file_cache;
	std::vector<directory *>	children;

	static const int CACHE_BUCKETS;
	static const char DB_FILE_NAME[];
	static const unsigned short  DB_VERSION;

private:

	// Found at the beginning of the dbfile.
	struct dbfile_header {
		char 		hstring[10];
		time_t 		timestamp;
		unsigned short 	version;
	};

	// Represents raw data of a file entry in the dbfile 
	struct raw_entry {
		off_t  offset;
		char 	* string_data;
		struct file_entry::header header;		
	};
	
	
public: 	
	directory(directory *, file_entry *);
	directory(const char *);
	~directory();

public:
	const char * abs_path() const  { return a_path; }
	const char * rel_path() const { return r_path; }
	const char * name() const  { return _name; }
	const char * owner() const { return _entry->owner(); }
	int flags() const { return _entry->flags(); }
	file_entry * entry() const { return _entry; }
	directory * parent() const { return _parent; }

	void inc_count() { _entry->inc_count(); }
	void dec_count() { _entry->dec_count(); }
	unsigned int usage_count() const { return _entry->usage_count(); }
	
	int batch_start();
	int batch_end();

	int create();
	int destroy();
	bool exists() const;
	int validate();
	int is_empty();
	
	file_entry * disk_lookup(const char *, int *);
	int write(const file_entry *, bool = false);
	int unlink(const char *, bool);
	int ls(const char *, std::vector<file_entry>&);

	template<typename Cont> void level_order_traverse(Cont&);

#ifdef __DEBUG__
	int dump(int);
#endif	

	enum {
		ERR_BASE             = -300,
		ERR_ILLEGAL_NAME     = -300,
		ERR_NOMATCH          = -301,
		ERR_READ             = -302,
		ERR_WRITE            = -303,
		ERR_BUSY             = -304,
		ERR_FAILURE          = -305,			
		ERR_DB_CORRUPTION    = -306,
		ERR_DB_VERSION       = -307,
		ERR_DB_OPEN          = -308,
		ERR_DB_EXISTS        = -309,
		ERR_DB_DESTROY       = -310,
		ERR_MAX              = -310
	};

	static const char * strerror(int);
	static bool is_legal_name(const char *);

public:
	/** 
	  * simple cache to keep track of file_entry objects created
	  */
	file_entry * cache_lookup(const char * name) const {	
		hash_table_t::const_iterator i = file_cache.find(name);
		return (i == file_cache.end()) ? NULL : (*i).second;
	}
	
	void cache_insert(file_entry * f) { 
		file_cache.insert(f->name(), f); 
	}

	void cache_remove(const char * name) {
		file_cache.erase(name);
	}

	int cache_num_open() const;
	int cache_prune();
	
private:
	/** implementation **/
	enum lookup_options {
		MATCH_LITERAL		= 0x0001,
		MATCH_WILDCARD		= 0x0002
	};

	void add_child(directory *);
	void remove_child(directory *);


	int opendb();
	int closedb();
	int read_header(dbfile_header *);
	int write_header(time_t);

	int find_free_entry(unsigned int, struct raw_entry *);
	int free_entry(off_t);
	int write_entry(const struct raw_entry *);
	int locate_entry(const char *, struct raw_entry *, int);

	static unsigned int calc_dynlen(const struct file_entry::header * h) {
		return (h->owner_length + h->from_length + 
			h->misc_length  + h->desc_length +
  					  h->name_length); 
	}	
	
private:	
	// non-copyable
	directory(const directory &);
	directory& operator=(const directory&);
};

/**
 * Return a level-order traversal of the active directory 
 * objects, starting from this one.
 */
template<typename Cont> void directory::level_order_traverse(Cont& out)
{
	std::deque<directory *> queue;
	queue.push_back(this);

	while (!queue.empty()) {
		directory * d = queue.front();
		out.push_back(d);

		std::copy(d->children.begin(), d->children.end(), std::back_inserter(queue));
		queue.pop_front();
	}
}

#ifdef __DEBUG__
/**
 * Assist with pretty printing of directory lists.
 */
inline std::ostream& operator << (std::ostream& out, const directory * d)
{
	out << "directory: '" << d->rel_path() << "'";
	return out;
}	
#endif

} // namespace fs
#endif
