/**
 * fs/library.h -- part of ezbounce
 * (c) 2003-2008 Murat Deligonul
 *
 * On-disk file library support.
 */

#ifndef __FS_LIBRARY_H
#define __FS_LIBRARY_H

#include <vector>
#include <cstring>
#include "util/hash.h"
#include "util/exception.h"
#include "util/counted_object.h"

// forward declarations
class userdef;

namespace fs {

class file_entry;
class directory;
class resolved_path;
class flib_key;

/**
 * Exception handling supprt 
 */
struct flib_exception_tag { };
typedef util::exception<flib_exception_tag> flib_exception;

class file_library {
private:
	typedef util::hash_table<const char *, directory *, 
			util::default_hasher<const char *>,
		       	util::fixed_cstring_predicate<strcmp> > hash_table_t;

	hash_table_t directories;
	directory * root;
	userdef * root_user;
	flib_key * root_key;

public:
	file_library(const char *);
	~file_library();
		
	flib_key * create_key(const userdef *, int *);	
	
	/** 'system' calls **/
	file_entry * open(const flib_key *, const char *, int, int *);	
	file_entry * import(const flib_key *, const char *);	
	int save(const flib_key *, file_entry *);
	int close(const flib_key *, file_entry *);
	
	int unlink(const flib_key *, const char *);	
	int access(const flib_key *, file_entry *, int);
	int access(const flib_key *, const char *, int);
		
	int chdir(flib_key *, const char *);	
	int rmdir(const flib_key *, const char *);	
	directory * mkdir(const flib_key *, const char *, int *);
	directory * mkdirs(const flib_key *, const char *, int *);

	int ls(const flib_key *, const char *, std::vector<file_entry>&);

	/** error codes **/
	enum {
		ERR_BASE	  = -200,
		ERR_BADFILE 	  = -200,
		ERR_CREATION 	  = -201,
		ERR_PRIVATE_FILE  = -202,
		ERR_ACCESS_READ   = -203,
		ERR_ACCESS_WRITE  = -204,
		ERR_DIRECTORY     = -206,
		ERR_NOT_DIRECTORY = -207,
		ERR_BUSY 	  = -208,
		ERR_FAILURE       = -209,
		ERR_OPEN_FILES    = -210,
		ERR_RMDIR_FAILURE = -211,
		ERR_MKDIR_FAILURE = -212,
		ERR_SYMLINK       = -213,
		ERR_ACCESS        = -214,
		ERR_EXISTS	  = -215,
		ERR_DOESNT_EXIST  = -216,
		ERR_EXISTS_ON_DISK= -217,
		ERR_CONFIG	  = -218,
		ERR_NO_KEY        = -219,
		ERR_NO_FILELIB    = -220,
	        ERR_MAX		  = -220
	};

	static const char * strerror(int);

private:
	/** implementation **/
	friend class resolved_path;

	int resolve_path(const flib_key *, const char *, resolved_path *);
	directory * load_directory(const flib_key *, directory *, const char *, int *);
	
	static int fixup_path(const flib_key * , const char *, char *);	

private:
	// non-copyable
	file_library& operator=(const file_library&);
	file_library(const file_library&);
};

class resolved_path {
private:
	char resolved[PATH_MAX+1];
	char * name_ptr;
	directory * _dir;

public:
	directory * dir() { return _dir; }
	char * buffer() { return resolved; }
	const char * full_path() const { return resolved; }
	const char * name() const { return name_ptr; }

	resolved_path() { 
		resolved[0] = 0;
		name_ptr = &resolved[0];
		_dir = 0;
	}

	friend int file_library::resolve_path(const flib_key *, const char *, resolved_path *);

private:
	// non-copyable
	resolved_path(const resolved_path&);
	resolved_path& operator= (const resolved_path&);
};

/**
  * Used by file library to grant access and check file permissions
  */
class flib_key : public util::counted_object<flib_key> {
private:
	friend class file_library;
	
	const userdef * const user;		/* user accessing files */
	directory * cwd;          		/* current working directory */

private:
	flib_key(const userdef * u, directory * d);

public:
	flib_key(const flib_key&);
	~flib_key();

	const char * username() const;
	const userdef * get_user() const { return user; }
	directory * dir() const { return cwd; }
	
        void set_dir(directory * d); 
	
private: 
	// non-copyable
	flib_key& operator = (const flib_key&);
};

} // namespace fs 
#endif
