/**
 * irc/addrcache.h
 * Part of ezbounce
 */
#ifndef __IRC_ADDRCACHE_H
#define __IRC_ADDRCACHE_H

#include "util/hash.h"
#include "irc/addrentry.h"
#include "net/socket.h"
#include "channel.h"
#include "rfunc.h"

#include "debug.h"

class textline;

namespace irc {

class server_info;

class address_cache {
public:
	/**
	  * Hash tables and their types.
	  */
	typedef util::hash_table<const char *, address_entry *, util::cstring_hash_fn, util::cstring_predicate> nick_hash_t;
	typedef util::hash_table<const char *, channel *, util::cstring_hash_fn, util::cstring_predicate> channel_hash_t;

private:
	nick_hash_t	nick_hash;
	channel_hash_t	channel_hash;
	const server_info * server;

public:
	address_cache(const server_info *, int, int);
	~address_cache();

	/** 
	 * Channel & nick management interface
	 * Return < 0 on error conditions
	 */
	int add_channel(const char *);			// user 'joins' a channel
	int delete_channel(const char *);		// user 'parts' a channel

	/**
	 * Lookup functions
	 */
	channel * lookup_channel(const char * name) { 
		channel_hash_t::iterator i = channel_hash.find(name);
		return (i == channel_hash.end()) ? NULL : (*i).second;
       	}
	address_entry * lookup(const char * nick) { 
		nick_hash_t::iterator i = nick_hash.find(nick);
		return (i == nick_hash.end()) ? NULL : (*i).second;
	}
	address_entry * lookup_nick(const char * n) {
	       	return lookup(n); 
	}

	size_t num_channels() const { 
		return channel_hash.size(); 
	}
	size_t num_nicks() const { 
		return nick_hash.size(); 
	}

	/**
	  * Iterators to channel list.
	  */
	channel_hash_t::const_iterator channels_begin() const {
		return channel_hash.begin();
	}

	channel_hash_t::const_iterator channels_end() const {
		return channel_hash.end();
	}

	int add_nick(const char * n, const char * addr = NULL, const char * chan = NULL);
	int delete_nick(const char *, const char * = NULL);
	int change_nick(const char *, const char *);
	int bind_nick(const char * nick, const char * addr);

	void set_nick_flags(const char * a, const char * b, server_info::mode_flag c) { flag_operator(1, a,b,c); }
	void clear_nick_flags(const char * a, const char * b, server_info::mode_flag c) { flag_operator(2, a,b,c); }
	int get_nick_flags(const char * a, const char * b) { return flag_operator(0, a, b, 0); }
	int flag_operator(int param, const char * nick, const char *, server_info::mode_flag);

	void stats(util::hash_stats *,	util::hash_stats *);

        /**
	 * Line parsing functions 
	 */
	int parse_MODE_line(const char * channel, const textline&);

	/**
	 * Block parsing functions
	 */
	int parse_NAMES_block(const char *, const textline&);
	int end_NAMES_block(const char *);

	int parse_WHO_block(const char *, const textline&);
	int end_WHO_block(const char *);

	enum rfunc_t {
		RFUNC_NAMES,
		RFUNC_WHO
	};

	rfunc * create_rfunc(rfunc_t, net::socket *, const char *, const char *);

private:
	int purge();
	
private:
	// noncopyable
	address_cache(const address_cache&);
	address_cache& operator=(const address_cache&);
};

} // namespace irc
#endif 
