/***************************************************************************
 * Part of ezbounce
 * (c) 2006-2008 Murat Deligonul
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 */

#include "autoconf.h"

#include <stdexcept>
#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "util/strings.h"
#include "fs/entry.h"
#include "fs/directory.h"

#include "debug.h"

using namespace util::strings;

// XXX:
// improvements:
// clean up constructors

namespace fs {

/**
  * In-memory representation of a file in the file library.
  * Obtained through file_library open() method.
  * Must be discarded through file_library close() method
  * File can be open()'ed for IO through file_entry::open()
  *		-however-
  *	1) only one writer is permitted at a time (io chain is stored
  *		as data member) 
  *     2) multiple simultaneous readers are permitted, even when
  *	   there is a writer
  * 	3) the 'usage_count' member variables keeps track of how many
  *  	   references to this object exist.
  *     4) upon completion of IO, call file_entry::close(int fd)  
  */
file_entry::file_entry(directory * d, const char * fname)
		: _dir(d),
		  _creation(time(NULL)),
		  _flags(0),
		  _usage_count(0),
		  _size(0),
		  fwriter(NULL)
{
	strings.set(NAME, fname);
	DEBUG("%s [%p] Constructed in directory [%p]: %s\n", __PRETTY_FUNCTION__, this, d, fname);	
}

file_entry::file_entry(directory * d, const header * h, const char * raw)
	: _dir(d),
		_usage_count(0),
		_size(0),
		fwriter(NULL)
{
	assert(h != 0);
	assert(raw);
	DEBUG("%s [%p]\n", __PRETTY_FUNCTION__, this);
	unsigned int offset = 0;
			
	set_owner(raw);
	offset += h->owner_length;

	if (h->from_length > 0) {
		set_from(&raw[offset]);
	}
	offset  += h->from_length;

	if (h->misc_length > 0) {
		set_misc(&raw[offset]);
	}
	offset += h->misc_length;

	if (h->desc_length > 0) {
		set_desc(&raw[offset]);
	}
	offset += h->desc_length;

	if (raw[offset]) {
		set_name(&raw[offset]);
	}

	_creation = h->creation;
	_flags	= h->flags;

	DEBUG("\t---> new entry: name: '%s' owner: %s\n", name(), owner());
	DEBUG("\t  \\--> pass: %s desc: %s from: %s\n", misc(), desc(), from());
	DEBUG("\t  \\--> creation: %ld flags: %d\n", _creation, _flags);	
}

/**
  * Copy constructor, called only w/ STL operations
  * FIXME: writer and usage_count
  */
file_entry::file_entry(const file_entry &that)
	: strings(that.strings) 
{
	DEBUG("%s: [%p] that: [%p]\n", __PRETTY_FUNCTION__, this, &that);
	_dir = that._dir;
	_creation = that._creation;
	_flags = that._flags;
	_size = that._size;
	// writer and usage_count aren't copied
	fwriter = NULL;
	_usage_count = 0;
}

/**
  * Operator =, called only w/ STL operations
  * FIXME: writer and usage_count
  *
  */
file_entry& file_entry::operator= (const file_entry& that)
{
	DEBUG("%s: [%p] that: [%p]\n", __PRETTY_FUNCTION__, this, &that);
	strings = that.strings;
	_dir = that._dir;
	_creation = that._creation;
	_flags = that._flags;
	_size = that._size;
	fwriter = that.fwriter;
	_usage_count = that._usage_count;
	return *this;
}

/**
  * Fill a header with what we can -- creation time, flags,
  * downloads, and the string data. 
  * Return length of the string data.
  */
unsigned int file_entry::fill_header(header * h) const
{
	assert(h);
	memset(h, 0, sizeof(header));

	unsigned int dyn = 0;
	h->creation  = creation();
	h->flags     = flags();

	if (is_non_empty(owner())) {
		dyn += (h->owner_length = strlen(owner()) + 1);
	}
	if (is_non_empty(misc())) {
		dyn += (h->misc_length = strlen(misc()) + 1);
	}
	if (is_non_empty(desc())) {
		dyn += (h->desc_length = strlen(desc()) + 1);
	}
	if (is_non_empty(name())) {
		dyn += (h->name_length = strlen(name()) + 1);
	}
	if (is_non_empty(from())) {
		dyn += (h->from_length = strlen(from()) + 1);
	}
	return dyn;
}

/**
 * Create raw data buffer to be written to database file.
 * The buffer must be as large as all of the length fields
 * combined. (which fill_header() will return) 
 */
int file_entry::fill_data(char * b) const
{
	int offset = 0;
	int len = 0;
	const char * table[] = {
		owner(), from(), misc(), desc(), name()
	};

	for (unsigned j = 0; j < 5; ++j) {
		len = (table[j] ? strlen(table[j]) : 0);
		if (len > 0) {
			memcpy(b + offset, table[j], len);
			*(b+offset + len) = 0;
			offset += len+1;
		}
	}
	return offset - 1;	
}

/**
  * Open file for reading.  Exception throw if error occurs.
  */
file_entry::reader_t * file_entry::open_reader(reader_t::openmode mode)
{
	const std::string& path = my_sprintf("%s/%s", dir()->abs_path(), name());
	reader_t * reader = new reader_t(path.c_str(), mode);
	update_size();
	return reader;
}

/**
 * Open file for writing.  Exception thrown if error occurs, or if file is 
 * already being written to.
 */
file_entry::writer_t * file_entry::open_writer(writer_t::openmode mode)
{
	if (fwriter != NULL) {
		throw std::logic_error("file is locked (currently being written to)");	
	}

	const std::string& path = my_sprintf("%s/%s", dir()->abs_path(), name());
	fwriter = new writer_t(path.c_str(), mode);
	update_size();
	DEBUG("file_entry::open(): Write access granted for %s\n", path.c_str());
	return fwriter;
}

/**
  * Close reader that has been opened for IO.  Also deletes the object; the
  * pointer provided will no longer be valid after this call.
  * NOTE: no way of checking whether the reader belonged to this file_entry.  It just
  * 	  deletes whatever reader it gets.  (Thus, NULL values are automatically ignored)
  */
void file_entry::close(reader_t * reader) 
{
	delete reader;
}

/**
 * Close writer that has been opened for IO.  The writer must belong to this file_entry,
 * but NULL values are ignored.
 * The writer object is also deleted; the pointer will no longer be valid after this call.
 */
void file_entry::close(writer_t * writer) 
{
	if (writer == NULL) {
		return;
	}
	if (writer != fwriter) {
		throw std::invalid_argument("writer doesn't belong to this file entry");
	}
	DEBUG("file_entry()::close(): Write access closed for %s/%s\n",
			dir()->abs_path(), name());
	delete fwriter;
	fwriter = NULL;	
}

/**
  * Uses stat(2) to determine size of the file we point to.
  * Set to 0 if file doesn't yet exist.
  */
void file_entry::update_size()
{
	struct stat st;
	memset(&st, 0, sizeof(st));
	char * path = my_strdup3(dir()->abs_path(), "/", name());
	int r = stat(path, &st);
	delete[] path;

	if (r >= 0) {
		set_size(st.st_size);
	} 
	else {
		set_size(0);
	}
}

/* static */ const char * file_entry::strerror(int err) 
{
	switch (err) {
	case ERR_LOCKED:
		return "File is locked (currently being written to)";
	case ERR_OPEN:
		return "Unable to open file for I/O";
	}
	return "Unknown error";
}

} // namespace fs
