/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * irc/channel.cc -- part of the ezbounce IRC proxy.
 * (C) 2007-2008 Murat Deligonul
 */

#include "autoconf.h"

#include <list>
#include "irc/channel.h"
#include "debug.h"

namespace irc {

using std::map;
using std::list;
using std::string;

/* static */ config::target_list channel::default_target_list(3, config::config_target("default"));

/**
 * Create a new channel object:
 * Copy the name, set logfile and flags to default values,
 * Copy network and server names into target list from server_info,
 * Initialize hash table with proper case-mapping specific functions.
 */
channel::channel(const server_info * s, const char * n, int i) :
	_name(my_strdup(n)),
	flags(0),
	log(NULL),
	logging(false),
	log_options(0),
	modes(0),
	channel_target_list( s->server_target() ),	
	nick_hash(i, s->nick_hash_func(), s->casecmp_func()) 
{ 
	nick_hash.set_rehash_trigger(0,3); 
	channel_target_list.push_back( config::config_target(_name) );
}

/**
 * NOTE: We don't remove references to us in the users
 *	 data structures. That's up to the caller.
 */
channel::~channel() 
{
	DEBUG("channel::~channel(): [%p] %s destroyed\n", this, _name);
	delete[] _name;
}

/**
  * Get a combined string of ALL channel modes.
  */
void channel::get_all_modes(const server_info * servinfo, std::string& out) const
{	
	map<char, string>::const_iterator i = valued_modes.begin(),
					e = valued_modes.end();
	list<string> 	mlist;

	mlist.push_back("+");
	string& front = mlist.front();

	for (; i != e; ++i) {
		front += (*i).first;
		mlist.push_back((*i).second);		
	}

	// now get regular modes
	char buffer[server_info::MAX_FLAGS + 1];
	servinfo->str_flags(server_info::CHAN_MODE_D, modes, buffer);
	front += buffer;

	// consolidate output into one string
	util::print_container(mlist, out, " ", NULL);
}

void channel::set_valued_mode(char c, const char * value)
{
	assert(c != 0);
	assert(value != NULL);

	map<char, string>::iterator i = valued_modes.find(c);
	if (i == valued_modes.end()) {
		valued_modes[c] = std::string(value);
	}
	else {
		(*i).second = value;
	}
}

} // namespace irc
